#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2022 Stéphane Caron
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Python inverse kinematics for your robot model based on Pinocchio.
"""

from .configuration import Configuration
from .configuration import assume_configuration
from .configuration import apply_configuration
from .solve_ik import solve_ik

__version__ = "0.2.0"

__all__ = [
    "Configuration",
    "apply_configuration",
    "assume_configuration",
    "models",
    "solve_ik",
]
